"""
MIT License

Copyright (c) 2021 Jedy Matt Tabasco

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from . import errors, util


class Key:
    def __init__(self, name: str, type_):
        self.name = name
        self.type_ = type_

    @classmethod
    def model(cls):
        return cls('model', str)

    @classmethod
    def data(cls):
        return cls('data', dict)

    @classmethod
    def filter(cls):
        return cls('filter', dict)

    def is_valid_type(self, entity):
        return isinstance(entity, self.type_)

    def __str__(self):
        return self.name

    def __eq__(self, o: object) -> bool:
        if isinstance(o, self.__class__):
            return self.name == o.name and self.type_ == o.type_

        if isinstance(o, str):
            return self.name == o

        return False

    def __hash__(self):
        return hash(self.name)


def check_model_key(entity: dict, entity_is_parent: bool):
    model = Key.model()
    if model not in entity and entity_is_parent:
        raise errors.MissingKeyError("'model' key is missing.")
    # check type_
    if model in entity and not model.is_valid_type(entity[model]):
        raise errors.InvalidTypeError("'model' data should be 'string'.")


def check_max_length(entity: dict):
    if len(entity) > 2:
        raise errors.MaxLengthExceededError("Length should not exceed by 2.")


def check_source_key(entity: dict, source_keys: list) -> Key:
    source_key: Key = next(
        (sk for sk in source_keys if sk in entity),
        None
    )

    # check if current keys has at least, data or filter key
    if source_key is None:
        raise errors.MissingKeyError(f"Missing {', '.join(map(str, source_keys))} key(s).")

    return source_key


def check_source_data(source_data, source_key: Key):
    if not isinstance(source_data, dict) and not isinstance(source_data, list):
        raise errors.InvalidTypeError(f"Invalid type_, {str(source_key)} should be either 'dict' or 'list'.")

    if isinstance(source_data, list) and len(source_data) == 0:
        raise errors.EmptyDataError("Empty list, 'data' or 'filter' list should not be empty.")


def check_data_type(item, source_key: Key):
    if not source_key.is_valid_type(item):
        raise errors.InvalidTypeError(
            f"Invalid type_, '{source_key.name}' should be '{source_key.type_}'")


class SchemaValidator:
    _source_keys = None
    _ref_prefix = None

    @classmethod
    def validate(cls, entities, ref_prefix='!', source_keys=None):
        if source_keys is None:
            cls._source_keys = [Key.data(), Key.filter()]
        cls._ref_prefix = ref_prefix

        cls._pre_validate(entities, entity_is_parent=True)

    @classmethod
    def _pre_validate(cls, entities: dict, entity_is_parent=True):
        if not isinstance(entities, dict) and not isinstance(entities, list):
            raise errors.InvalidTypeError("Invalid type, should be list or dict")
        if len(entities) == 0:
            return
        if isinstance(entities, dict):
            return cls._validate(entities, entity_is_parent)
        # iterate list
        for entity in entities:
            cls._pre_validate(entity, entity_is_parent)

    @classmethod
    def _validate(cls, entity: dict, entity_is_parent=True):
        check_max_length(entity)
        check_model_key(entity, entity_is_parent)

        # get source key, either data or filter key
        source_key = check_source_key(entity, cls._source_keys)
        source_data = entity[source_key]

        check_source_data(source_data, source_key)

        if isinstance(source_data, list):
            for item in source_data:
                check_data_type(item, source_key)
                # check if item is a relationship attribute
                cls.check_attributes(item)
        else:
            # source_data is dict
            # check if item is a relationship attribute
            cls.check_attributes(source_data)

    @classmethod
    def check_attributes(cls, source_data: dict):
        for _, value in util.iter_ref_kwargs(source_data, cls._ref_prefix):
            cls._pre_validate(value, entity_is_parent=False)



def iter_ref_kwargs(kwargs: dict, ref_prefix: str):
    """Iterate kwargs with name prefix or references"""
    for attr_name, value in kwargs.items():
        if attr_name.startswith(ref_prefix):
            # removed prefix
            yield attr_name[len(ref_prefix):], value


def iter_non_ref_kwargs(kwargs: dict, ref_prefix: str):
    """Iterate kwargs, skipping item with name prefix or references"""
    for attr_name, value in kwargs.items():
        if not attr_name.startswith(ref_prefix):
            yield attr_name, value



"""
MIT License

Copyright (c) 2021 Jedy Matt Tabasco

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import csv
import json
import sys

from . import validator

try:
    import yaml
except ModuleNotFoundError:
    pass


def load_entities_from_json(json_filepath):
    try:
        with open(json_filepath, 'r') as f:
            entities = json.loads(f.read())
    except FileNotFoundError as error:
        raise FileNotFoundError(error)

    validator.SchemaValidator.validate(entities)

    return entities


def load_entities_from_yaml(yaml_filepath):
    if 'yaml' not in sys.modules:
        raise ModuleNotFoundError(
            'PyYAML is not installed and is required to run this function. '
            'To use this function, py -m pip install "sqlalchemyseed[yaml]"'
        )

    try:
        with open(yaml_filepath, 'r') as f:
            entities = yaml.load(f.read(), Loader=yaml.SafeLoader)
    except FileNotFoundError as error:
        raise FileNotFoundError(error)

    validator.SchemaValidator.validate(entities)

    return entities


def load_entities_from_csv(csv_filepath: str, model) -> dict:
    """Load entities from csv file

    :param csv_filepath: string csv file path
    :param model: either str or class
    :return: dict of entities
    """
    with open(csv_filepath, 'r') as f:
        source_data = list(map(dict, csv.DictReader(f, skipinitialspace=True)))
        if isinstance(model, str):
            model_name = model
        else:
            model_name = '.'.join([model.__module__, model.__name__])

        entities = {'model': model_name, 'data': source_data}

    validator.SchemaValidator.validate(entities)

    return entities



"""
MIT License

Copyright (c) 2021 Jedy Matt Tabasco

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import importlib
from sqlalchemy.exc import NoInspectionAvailable
from inspect import isclass
from sqlalchemy import inspect
from . import errors


def parse_class_path(class_path: str):
    try:
        module_name, class_name = class_path.rsplit('.', 1)
    except ValueError:
        raise errors.ParseError('Invalid module or class input format.')

    # if class_name not in classes:
    class_ = getattr(importlib.import_module(module_name), class_name)

    try:
        if isclass(class_) and inspect(class_):
            return class_
        else:
            raise TypeError("'{}' is not a class".format(class_name))
    except NoInspectionAvailable:
        raise TypeError(
            "'{}' is an unsupported class".format(class_name))

class ClassRegistry:
    def __init__(self):
        self._classes = {}

    def register_class(self, class_path: str):
        """

        :param class_path: module.class (str)
        :return: registered class
        """

        if class_path not in self._classes:
            self._classes[class_path] = parse_class_path(class_path)

        return self._classes[class_path]

    def __getitem__(self, class_path: str):
        return self._classes[class_path]

    @property
    def classes(self):
        return tuple(self._classes)

    def clear(self):
        self._classes.clear()



class ClassNotFoundError(Exception):
    """Raised when the class is not found"""
    pass


class MissingKeyError(Exception):
    """Raised when a required key is missing"""
    pass


class MaxLengthExceededError(Exception):
    """Raised when maximum length of data exceeded"""
    pass


class InvalidTypeError(Exception):
    """Raised when a type of data is not accepted"""
    pass


class EmptyDataError(Exception):
    """Raised when data is empty"""
    pass


class InvalidKeyError(Exception):
    """Raised when an invalid key is invoked"""
    pass

class ParseError(Exception):
    """Raised when parsing string fails"""
    pass


import os
import re

from setuptools import setup


# with open(os.path.join('src', 'sqlalchemyseed', '__init__.py'), 'r') as f:
#     pattern = r"^__version__ = ['\"]([^'\"]*)['\"]"
#     VERSION = re.search(pattern, f.read(), re.MULTILINE).group(1)


# setup(version=VERSION,)
setup()



"""
MIT License

Copyright (c) 2021 Jedy Matt Tabasco

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import abc
from typing import NamedTuple

import sqlalchemy
from sqlalchemy.orm import ColumnProperty
from sqlalchemy.orm import object_mapper
from sqlalchemy.orm.relationships import RelationshipProperty
from sqlalchemy.sql import schema

from . import class_registry, validator, errors, util


class AbstractSeeder(abc.ABC):

    @property
    @abc.abstractmethod
    def instances(self, *args, **kwargs):
        pass

    @abc.abstractmethod
    def seed(self, *args, **kwargs):
        pass

    @abc.abstractmethod
    def _pre_seed(self, *args, **kwargs):
        pass

    @abc.abstractmethod
    def _seed(self, *args, **kwargs):
        pass

    @abc.abstractmethod
    def _seed_children(self, *args, **kwargs):
        pass

    @abc.abstractmethod
    def _setup_instance(self, *args, **kwargs):
        pass


class EntityTuple(NamedTuple):
    instance: object
    attr_name: str


class Entity(EntityTuple):
    @property
    def class_attribute(self):
        return getattr(self.instance.__class__, self.attr_name)

    @property
    def instance_attribute(self):
        return getattr(self.instance, self.attr_name)

    @instance_attribute.setter
    def instance_attribute(self, value):
        setattr(self.instance, self.attr_name, value)

    def is_column_attribute(self):
        return isinstance(self.class_attribute.property, ColumnProperty)

    def is_relationship_attribute(self):
        return isinstance(self.class_attribute.property, RelationshipProperty)

    @property
    def referenced_class(self):
        # if self.is_column_attribute():
        #     return
        if self.is_relationship_attribute():
            return self.class_attribute.mapper.class_

        if self.is_column_attribute():
            table_name = get_foreign_key_column(self.class_attribute).table.name
            return next(
                (
                    mapper.class_
                    for mapper in object_mapper(self.instance).registry.mappers
                    if mapper.class_.__tablename__ == table_name
                ),
                errors.ClassNotFoundError(
                    "A class with table name '{}' is not found in the mappers".format(table_name)
                )
            )


def get_foreign_key_column(attr, idx=0) -> schema.Column:
    return list(attr.foreign_keys)[idx].column


def filter_kwargs(kwargs: dict, class_, ref_prefix):
    return {
        k: v for k, v in util.iter_non_ref_kwargs(kwargs, ref_prefix)
        if not isinstance(getattr(class_, str(k)).property, RelationshipProperty)
    }


def set_parent_attr_value(instance, parent: Entity):
    if parent.is_relationship_attribute():
        if parent.class_attribute.property.uselist is True:
            parent.instance_attribute.append(instance)
        else:
            parent.instance_attribute = instance

    if parent.is_column_attribute():
        parent.instance_attribute = instance


class Seeder(AbstractSeeder):
    __model_key = validator.Key.model()
    __data_key = validator.Key.data()

    def __init__(self, session: sqlalchemy.orm.Session = None, ref_prefix="!"):
        self.session = session
        self._class_registry = class_registry.ClassRegistry()
        self._instances = []
        self.ref_prefix = ref_prefix

    @property
    def instances(self):
        return tuple(self._instances)

    def get_model_class(self, entity, parent: Entity):
        if self.__model_key in entity:
            return self._class_registry.register_class(entity[self.__model_key])
        # parent is not None
        return parent.referenced_class

    def seed(self, entities, add_to_session=True):
        validator.SchemaValidator.validate(
            entities, ref_prefix=self.ref_prefix, source_keys=[validator.Key.data()])

        self._instances.clear()
        self._class_registry.clear()

        self._pre_seed(entities)

        if add_to_session:
            self.session.add_all(self.instances)

    def _pre_seed(self, entity, parent: Entity = None):
        if isinstance(entity, dict):
            self._seed(entity, parent)
        else:  # is list
            for item in entity:
                self._pre_seed(item, parent)

    def _seed(self, entity, parent: Entity = None):
        class_ = self.get_model_class(entity, parent)

        kwargs = entity[self.__data_key]

        # kwargs is list
        if isinstance(kwargs, list):
            for kwargs_ in kwargs:
                instance = self._setup_instance(class_, kwargs_, parent)
                self._seed_children(instance, kwargs_)
            return

        # kwargs is dict
        # instantiate object
        instance = self._setup_instance(class_, kwargs, parent)
        self._seed_children(instance, kwargs)

    def _seed_children(self, instance, kwargs):
        for attr_name, value in util.iter_ref_kwargs(kwargs, self.ref_prefix):
            self._pre_seed(entity=value, parent=Entity(instance, attr_name))

    def _setup_instance(self, class_, kwargs: dict, parent: Entity):
        instance = class_(**filter_kwargs(kwargs, class_, self.ref_prefix))
        if parent is not None:
            set_parent_attr_value(instance, parent)
        else:
            self._instances.append(instance)
        return instance

    # def instantiate_class(self, class_, kwargs: dict, key: validator.Key):
    #     filtered_kwargs = {
    #         k: v
    #         for k, v in kwargs.items()
    #         if not k.startswith("!")
    #            and not isinstance(getattr(class_, k), RelationshipProperty)
    #     }
    #
    #     if key is validator.Key.data():
    #         return class_(**filtered_kwargs)
    #
    #     if key is validator.Key.filter() and self.session is not None:
    #         return self.session.query(class_).filter_by(**filtered_kwargs).one()


class HybridSeeder(AbstractSeeder):
    __model_key = validator.Key.model()
    __source_keys = [validator.Key.data(), validator.Key.filter()]

    def __init__(self, session: sqlalchemy.orm.Session, ref_prefix: str = '!'):
        self.session = session
        self._class_registry = class_registry.ClassRegistry()
        self._instances = []
        self.ref_prefix = ref_prefix

    @property
    def instances(self):
        return tuple(self._instances)

    def get_model_class(self, entity, parent: Entity):
        # if self.__model_key in entity and (parent is not None and parent.is_column_attribute()):
        #     raise errors.InvalidKeyError("column attribute does not accept 'model' key")

        if self.__model_key in entity:
            class_path = entity[self.__model_key]
            return self._class_registry.register_class(class_path)

        # parent is not None
        if parent is not None:
            return parent.referenced_class

    def seed(self, entities):
        validator.SchemaValidator.validate(
            entities, ref_prefix=self.ref_prefix)

        self._instances.clear()
        self._class_registry.clear()

        self._pre_seed(entities)

    def _pre_seed(self, entity, parent=None):
        if isinstance(entity, dict):
            self._seed(entity, parent)
        else:  # is list
            for item in entity:
                self._pre_seed(item, parent)

    def _seed(self, entity, parent):
        class_ = self.get_model_class(entity, parent)

        source_key: validator.Key = next(
            (sk for sk in self.__source_keys if sk in entity),
            None
        )

        source_data = entity[source_key]

        # source_data is list
        if isinstance(source_data, list):
            for kwargs in source_data:
                instance = self._setup_instance(class_, kwargs, source_key, parent)
                self._seed_children(instance, kwargs)
            return

        # source_data is dict
        instance = self._setup_instance(class_, source_data, source_key, parent)
        self._seed_children(instance, source_data)

    def _seed_children(self, instance, kwargs):
        for attr_name, value in util.iter_ref_kwargs(kwargs, self.ref_prefix):
            self._pre_seed(entity=value, parent=Entity(instance, attr_name))

    def _setup_instance(self, class_, kwargs: dict, key, parent):
        filtered_kwargs = filter_kwargs(kwargs, class_, self.ref_prefix)

        if key == key.data():
            instance = self._setup_data_instance(class_, filtered_kwargs, parent)
        else:  # key == key.filter()
            # instance = self.session.query(class_).filter_by(**filtered_kwargs)
            instance = self._setup_filter_instance(class_, filtered_kwargs, parent)

        # setting parent
        if parent is not None:
            set_parent_attr_value(instance, parent)

        return instance

    def _setup_data_instance(self, class_, filtered_kwargs, parent: Entity):
        if parent is not None and parent.is_column_attribute():
            raise errors.InvalidKeyError("'data' key is invalid for a column attribute.")

        instance = class_(**filtered_kwargs)
        self.session.add(instance)
        if parent is None:
            self._instances.append(instance)

        return instance

    def _setup_filter_instance(self, class_, filtered_kwargs, parent: Entity):
        if parent is not None and parent.is_column_attribute():
            foreign_key_column = get_foreign_key_column(parent.class_attribute)
            return self.session.query(foreign_key_column).filter_by(**filtered_kwargs).one()[0]

        if parent is not None and parent.is_relationship_attribute():
            return self.session.query(parent.referenced_class).filter_by(**filtered_kwargs).one()

        return self.session.query(class_).filter_by(**filtered_kwargs).one()
